package be.iminds.ilabt.jfed.lowlevel.api;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.lowlevel.resourceid.ResourceId;
import be.iminds.ilabt.jfed.lowlevel.resourceid.ResourceUrn;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.util.*;

/**
 * <p>This is a simple wrapper for the methods specified in the {@link <a href="http://www.protogeni.net/ProtoGeni/wiki/SliceAuthorityAPI">ProtoGeni Slice Authority API</a>}.
 * <p>All functionality provided by the API is available in this implementation, and little processing of results is done.
 * This class stores no internal state, except for the logger used. It sends all results and debug info to this {@link Logger} if it is non null.
 * <p/>
 * <p>All methods require a {@link be.iminds.ilabt.jfed.lowlevel.connection.SfaSslWithClientAuthConnection} argument and return a {@link ProtogeniSliceAuthority.SliceAuthorityReply} which wraps the actual return value of the call, if any.
 *
 * @author Wim Van de Meerssche
 * @see <a href="http://www.protogeni.net/ProtoGeni/wiki/SliceAuthorityAPI">ProtoGeni Slice Authority API</a>
 */
public class ProtogeniSliceAuthority extends AbstractApi {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ProtogeniSliceAuthority.class);


    /**
     * The protogeni SA "credentials" argument, takes an array with credential string as argument.
     * So NOT an array with hashes like Geni AM methods
     */
    @SuppressWarnings("unchecked")
    public static Vector createCredentialsVector(List<AnyCredential> credentialList) {
        Vector credentialsVector = new Vector();
        if (credentialList == null || credentialList.isEmpty()) return credentialsVector;

        credentialsVector.setSize(credentialList.size());
        for (int i = 0; i < credentialList.size(); i++) {
            AnyCredential c = credentialList.get(i);
            if (c == null)
                throw new IllegalArgumentException("Illegal arguments: no AnyCredential is allowed to be null in credentialList: " + credentialList);
            assert c.getCredentialXml() != null;
            credentialsVector.set(i, c.getCredentialXml());
        }
        return credentialsVector;
    }

    /**
     * Construct a new SliceAuthority, with a certain logger.
     * <p/>
     * <p>This class stores no internal state, except for the logger used. It sends all results and debug info to this {@link Logger} if it is non null.
     *
     * @param logger        the logger to use. May be null if no logger is used.
     * @param retrySettings the retrySettings to use, or null to use the static globalRetrySettings in AbstractApi
     */
    public ProtogeniSliceAuthority(@Nonnull Logger logger, RetrySettings retrySettings, @Nonnull JFedPreferences jFedPreferences) {
        super(logger, retrySettings, new ApiInfo.Api(ApiInfo.ApiName.PROTOGENI_SA, 1), jFedPreferences);
    }

    /**
     * Construct a new SliceAuthority, with a certain logger.
     * <p/>
     * <p>This class stores no internal state, except for the logger used. It sends all results and debug info to this {@link Logger} if it is non null.
     * <p/>
     * <p>This constructor sets autoRetryBusy to true, so "busy" replies are retried.
     *
     * @param logger the logger to use. May be null if no logger is used.
     */
    public ProtogeniSliceAuthority(@Nonnull Logger logger, @Nonnull JFedPreferences jFedPreferences) {
        super(logger, jFedPreferences.getRetrySettings(), new ApiInfo.Api(ApiInfo.ApiName.PROTOGENI_SA, 1), jFedPreferences);
    }

    /**
     * A human readable name for the implemented API
     *
     * @return "ProtoGeni Slice Authority API v1"
     */
    static public String getApiName() {
        return "ProtoGeni Slice Authority API v1";
    }

    /**
     * A human readable name for the implemented API
     *
     * @return "ProtoGeni Slice Authority API v1"
     */
    @Override
    public String getName() {
        return getApiName();
    }

    @Override
    protected boolean isBusyReply(be.iminds.ilabt.jfed.lowlevel.XMLRPCCallDetails res) {
        if (res instanceof XMLRPCCallDetailsWithCodeValueError) {
            XMLRPCCallDetailsWithCodeValueError geniDetails = (XMLRPCCallDetailsWithCodeValueError) res;
            int code = (Integer) geniDetails.getResultCode();
            return GeniAMResponseCode.getByCode(code).isBusy();
        }
        return false;
    }

    /**
     * GetVersion call: Get info about the version of the API supported at the server.
     *
     * @param con the {@link be.iminds.ilabt.jfed.lowlevel.connection.SfaSslWithClientAuthConnection} to use for this call. Must be a connection to a slice authority.
     * @return The version as a String, wrapped in a {@code SliceAuthority.SliceAuthorityReply}
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    @ApiMethod(order = 1, hint = "GetVersion call: Get info about the version of the API supported at the server.")
    public SliceAuthorityReply<String> getVersion(SfaConnection con) throws JFedException {
        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "GetVersion", new Vector(), null);
        SliceAuthorityReply<String> r = new SliceAuthorityReply<String>(res);
        log(res, r, "getVersion", "GetVersion", con, null);
        return r;
    }

    /**
     * GetCredential call, without arguments: Get the credential of the current user (= the user used to authorize the {@link be.iminds.ilabt.jfed.lowlevel.connection.SfaSslWithClientAuthConnection}).
     *
     * @param con the {@code GeniConnection} to use for this call. Must be a connection to a slice authority.
     * @return the {@code AnyCredential} for the current user, wrapped in a {@code SliceAuthority.SliceAuthorityReply}
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    @ApiMethod(order = 2, hint = "GetCredential call, without arguments: Get the credential of the current user.")
    public SliceAuthorityReply<AnyCredential> getCredential(SfaConnection con) throws JFedException {
        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "GetCredential", new Vector(), null);
        SliceAuthorityReply<AnyCredential> r = null;

        int intCode = (Integer) res.getResultCode();
        GeniResponseCode genicode = GeniAMResponseCode.getByCode(intCode);
        if (genicode.isSuccess()) {
            try {
                AnyCredential signedCredential =
                        AnyCredential.createSfa2("SliceAuthority getCredential", res.getResultValueObject().toString());
                r = new SliceAuthorityReply<AnyCredential>(res, signedCredential);
            } catch (Exception e) {
                log(res, null, "getCredential", "GetCredential", con, null);
                throw new JFedException("Exception retrieving Credential for getCredential call.", e, res, genicode);
            }
        } else {
            r = new SliceAuthorityReply<AnyCredential>(res, null);
        }
        log(res, r, "getCredential", "GetCredential", con, null);
        return r;

    }

    /**
     * GetCredential call, with a slice ID arguments: Get the credential of the specified slice.
     *
     * @param con            the {@code GeniConnection} to use for this call. Must be a connection to a slice authority.
     * @param userCredential the credential for the user requesting the slice credential
     * @param slice          a slice ID
     * @return the {@code AnyCredential } for the requested slice, wrapped in a {@code SliceAuthority.SliceAuthorityReply}
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    @ApiMethod(order = 3, hint = "GetCredential call, with a slice ID arguments: Get the credential of the specified slice.")
    public SliceAuthorityReply<AnyCredential> getSliceCredential(SfaConnection con,
                                                                 @ApiMethodParameter(name = "userCredential", hint = "the credential for the user requesting the slice credential")
                                                                 AnyCredential userCredential,
                                                                 @ApiMethodParameter(name = "slice", hint = "a slice ID (typically the slice URN).")
                                                                 ResourceId slice) throws JFedException {
        return getSliceCredential(con, userCredential.toCredentialList(), true, slice);
    }

    @ApiMethod(order = 4, hint = "GetCredential call, with a slice ID arguments: Get the credential of the specified slice.")
    public SliceAuthorityReply<AnyCredential> getSliceCredential(SfaConnection con,
                                                                 @ApiMethodParameter(name = "userCredentials", hint = "a list of credentials for the user requesting the slice credential")
                                                                 List<AnyCredential> userCredentials,
                                                                 @ApiMethodParameter(name = "slice", hint = "a slice ID (typically the slice URN).")
                                                                 ResourceId slice) throws JFedException {
        return getSliceCredential(con, userCredentials, false, slice);
    }
    protected SliceAuthorityReply<AnyCredential> getSliceCredential(SfaConnection con,
                                                                    List<AnyCredential> userCredentials,
                                                                    boolean forceSendSingleCredential /*if true, a single credential is sent instead of a list*/,
                                                                    ResourceId slice) throws JFedException {
        Map<String, Object> methodParams;
        if (forceSendSingleCredential) {
            assert !userCredentials.isEmpty();
            methodParams = makeMethodParameters("userCredential", userCredentials.get(0), "slice", slice);
        }
        else
            methodParams = makeMethodParameters("userCredentials", userCredentials, "slice", slice);

        Hashtable params = new Hashtable();

        params.put(slice.getType(), slice.getValue());
        params.put("type", "Slice");
        if (userCredentials != null) {
            if (forceSendSingleCredential) {
                params.put("credential", userCredentials.get(0).getCredentialXml());
            } else {
                Vector credentials = createCredentialsVector(userCredentials);
                params.put("credentials", credentials);
            }
        }

        Vector argv = new Vector();
        argv.setSize(1);
        argv.set(0, params);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "GetCredential", argv, methodParams);

        int intCode = (Integer) res.getResultCode();
        GeniResponseCode genicode = GeniAMResponseCode.getByCode(intCode);

        SliceAuthorityReply<AnyCredential> r = null;
        if (genicode.isSuccess())
            try {
                AnyCredential sliceCredential =
                        AnyCredential.createSfa2("SliceAuthority getSliceCredential", res.getResultValueObject().toString());
                r = new SliceAuthorityReply<AnyCredential>(res, sliceCredential);
            } catch (Exception e) {
                log(res, null, "getSliceCredential", "GetCredential", con, methodParams);
                throw new JFedException("Error parsing getSliceCredential result credential:" + e.getMessage(), e, res, genicode);
            }
        else {
            r = new SliceAuthorityReply<AnyCredential>(res, null);
        }
        log(res, r, "getSliceCredential", "GetCredential", con, methodParams);
        return r;
    }

    /**
     * GetCredential call, with any type of argument: Get the credential of the specified type.
     *
     * @param con            the {@code GeniConnection} to use for this call. Must be a connection to a slice authority.
     * @param userCredential the credential for the user requesting the slice credential
     * @param type           the type of object for which to request a credential
     * @param object          a slice ID
     * @return the {@code AnyCredential } for the requested slice, wrapped in a {@code SliceAuthority.SliceAuthorityReply}
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    @ApiMethod(order = 5, hint = "GetCredential call, with any type: Get the credential for an object of the specified type.")
    public SliceAuthorityReply<AnyCredential> getAnyCredential(SfaConnection con,
                                                               @ApiMethodParameter(name = "userCredential", hint = "the credential for the user requesting the slice credential")
                                                               AnyCredential userCredential,
                                                               @ApiMethodParameter(name = "type", hint = "the type of object for which to request a credential (for example \"Slice\")", guiDefault = "Slice", required = false)
                                                               String type,
                                                               @ApiMethodParameter(name = "resource", hint = "an ID (typically an URN or UUID).", required = false)
                                                               ResourceId object) throws JFedException {

        return getAnyCredential(con, userCredential.toCredentialList(), true, type, object);
    }
    @ApiMethod(order = 6, hint = "GetCredential call, with any type: Get the credential for an object of the specified type.")
    public SliceAuthorityReply<AnyCredential> getAnyCredential(SfaConnection con,
                                                               @ApiMethodParameter(name = "userCredentials",
                                                                       hint = "a list of credentials for the user requesting a credential",
                                                                       parameterType = ApiMethodParameterType.LIST_OF_CREDENTIAL)
                                                               List<AnyCredential> userCredentials,
                                                               @ApiMethodParameter(name = "type", hint = "the type of object for which to request a credential (for example \"Slice\")", guiDefault = "Slice", required = false)
                                                               String type,
                                                               @ApiMethodParameter(name = "object", hint = "an ID (typically an URN or UUID).", required = false)
                                                               ResourceId object) throws JFedException {
        return getAnyCredential(con, userCredentials, false, type, object);
    }


    public SliceAuthorityReply<AnyCredential> getAnyCredential(SfaConnection con,
                                                               List<AnyCredential> userCredentials,
                                                               boolean forceSendSingleCredential /*if true, a single credential is sent instead of a list*/,
                                                               String type,
                                                               ResourceId object) throws JFedException {
        Map<String, Object> methodParams;
        if (forceSendSingleCredential) {
            assert !userCredentials.isEmpty();
            methodParams = makeMethodParameters("userCredential", userCredentials.get(0), "type", type, "object", object);
        } else {
            methodParams = makeMethodParameters("userCredentials", userCredentials, "type", type, "object", object);
        }

        Hashtable params = new Hashtable();

        if (object != null)
            params.put(object.getType()/*example type: "urn" */, object.getValue());
        if (type != null)
            params.put("type", type);
        if (userCredentials != null) {
            if (forceSendSingleCredential) {
                params.put("credential", userCredentials.get(0).getCredentialXml());
            } else {
                Vector credentials = createCredentialsVector(userCredentials);
                params.put("credentials", credentials);
            }
        }

        Vector argv = new Vector();
        argv.setSize(1);
        argv.set(0, params);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "GetCredential", argv, methodParams);

        int intCode = (Integer) res.getResultCode();
        GeniResponseCode genicode = GeniAMResponseCode.getByCode(intCode);

        SliceAuthorityReply<AnyCredential> r = null;
        if (genicode.isSuccess())
            try {
                AnyCredential sliceCredential =
                        AnyCredential.createSfa2("SliceAuthority getAnyCredential for type " + type, res.getResultValueObject().toString());
                r = new SliceAuthorityReply<AnyCredential>(res, sliceCredential);
            } catch (Exception e) {
                log(res, null, "getAnyCredential", "GetCredential", con, methodParams);
                throw new JFedException("Error parsing getSliceCredential result credential:" + e.getMessage(), e, res, genicode);
            }
        else {
            r = new SliceAuthorityReply<AnyCredential>(res, null);
        }
        log(res, r, "getAnyCredential", "GetCredential", con, methodParams);
        return r;
    }

    /**
     * Resolve call, with a slice ID arguments: Get info about the specified slice.
     *
     * @param con            the {@code GeniConnection} to use for this call. Must be a connection to a slice authority.
     * @param userCredential the credential for the user requesting the info
     * @param slice          a slice ID
     * @return a Hashtable containing info about the slice, wrapped in a {@code SliceAuthority.SliceAuthorityReply}
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    @ApiMethod(order = 10, hint = "Resolve call, with a slice ID arguments: Get info about the specified slice.")
    public SliceAuthorityReply<Hashtable> resolveSlice(SfaConnection con,
                                                       @ApiMethodParameter(name = "userCredential", hint = "the credential for the user requesting the info")
                                                       AnyCredential userCredential,
                                                       @ApiMethodParameter(name = "slice", hint = "a slice ID (typically the slice URN).")
                                                       ResourceId slice) throws JFedException {
        assert con != null;
        assert userCredential != null;

        return resolveSlice(con, userCredential.toCredentialList(), true, slice);
    }
    @ApiMethod(order = 11, hint = "Resolve call, with a slice ID arguments: Get info about the specified slice.")
    public SliceAuthorityReply<Hashtable> resolveSlice(SfaConnection con,
                                                       @ApiMethodParameter(name = "userCredentials", hint = "a list of credentials for the user requesting the info")
                                                       List<AnyCredential> userCredentials,
                                                       @ApiMethodParameter(name = "slice", hint = "a slice ID (typically the slice URN).")
                                                       ResourceId slice) throws JFedException {
        assert con != null;
        assert userCredentials != null;

        return resolveSlice(con, userCredentials, false, slice);
    }

    public SliceAuthorityReply<Hashtable> resolveSlice(SfaConnection con,
                                                       List<AnyCredential> userCredentials,
                                                       boolean forceSendSingleCredential /*if true, a single credential is sent instead of a list*/,
                                                       ResourceId slice) throws JFedException {
        assert con != null;
        assert userCredentials != null;

        Map<String, Object> methodParams;
        if (forceSendSingleCredential) {
            assert !userCredentials.isEmpty();
            methodParams = makeMethodParameters("userCredential", userCredentials.get(0), "slice", slice);
        } else {
            methodParams = makeMethodParameters("userCredentials", userCredentials, "slice", slice);
        }

        Hashtable params = new Hashtable();
        params.put(slice.getType(), slice.getValue());
        if (forceSendSingleCredential) {
            params.put("credential", userCredentials.get(0).getCredentialXml());
        } else {
            Vector credentials = createCredentialsVector(userCredentials);
            params.put("credentials", credentials);
        }
        params.put("type", "Slice");

        Vector argv = new Vector();
        argv.setSize(1);
        argv.set(0, params);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "Resolve", argv, methodParams);

        SliceAuthorityReply<Hashtable> r = null;

        Object resVal = res.getResultValueObject();
        if (!(resVal instanceof Hashtable)) {
            r = new SliceAuthorityReply<Hashtable>(res, null);
        } else {
            Hashtable sliceTable = (Hashtable) resVal;
            r = new SliceAuthorityReply<Hashtable>(res, sliceTable);
        }
        log(res, r, "resolveSlice", "Resolve", con, methodParams);
        return r;
    }

    /**
     * Resolve call, with a user ID arguments: Get info about the specified user.
     *
     * @param con            the {@code GeniConnection} to use for this call. Must be a connection to a slice authority.
     * @param userCredential the credential for the user requesting the info
     * @param user           a user ID
     * @return a Hashtable containing info about the user, wrapped in a {@code SliceAuthority.SliceAuthorityReply}
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    @ApiMethod(order = 12, hint = "Resolve call, with a user ID arguments: Get info about the specified user.")
    public SliceAuthorityReply<Hashtable> resolveUser(SfaConnection con,
                                                      @ApiMethodParameter(name = "userCredential", hint = "the credential for the user requesting the info")
                                                      AnyCredential userCredential,
                                                      @ApiMethodParameter(name = "user", hint = "a user ID (typically the user URN).")
                                                      ResourceId user) throws JFedException {
        assert con != null;
        assert userCredential != null;

        return resolveUser(con, userCredential.toCredentialList(), true, user);
    }
    @ApiMethod(order = 13, hint = "Resolve call, with a user ID arguments: Get info about the specified user.")
    public SliceAuthorityReply<Hashtable> resolveUser(SfaConnection con,
                                                      @ApiMethodParameter(name = "userCredentials",
                                                              hint = "a list of credentials for the user requesting the info",
                                                              parameterType = ApiMethodParameterType.LIST_OF_CREDENTIAL)
                                                      List<AnyCredential> userCredentials,
                                                      @ApiMethodParameter(name = "user", hint = "a user ID (typically the user URN).")
                                                      ResourceId user) throws JFedException {
        assert con != null;
        assert userCredentials != null;

        return resolveUser(con, userCredentials, false, user);
    }

    public SliceAuthorityReply<Hashtable> resolveUser(SfaConnection con,
                                                      List<AnyCredential> userCredentials,
                                                      boolean forceSendSingleCredential /*if true, a single credential is sent instead of a list*/,
                                                      ResourceId user) throws JFedException {
        assert con != null;
        assert userCredentials != null;

        Map<String, Object> methodParams;
        if (forceSendSingleCredential) {
            assert !userCredentials.isEmpty();
            methodParams = makeMethodParameters("userCredential", userCredentials.get(0), "user", user);
        } else {
            methodParams = makeMethodParameters("userCredentials", userCredentials, "user", user);
        }

        Hashtable params = new Hashtable();
        params.put(user.getType(), user.getValue());
        if (forceSendSingleCredential) {
            params.put("credential", userCredentials.get(0).getCredentialXml());
        } else {
            Vector credentials = createCredentialsVector(userCredentials);
            params.put("credentials", credentials);
        }
        params.put("type", "User");

        Vector argv = new Vector();
        argv.setSize(1);
        argv.set(0, params);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "Resolve", argv, methodParams);

        SliceAuthorityReply<Hashtable> r = null;

        Object resVal = res.getResultValueObject();
        if (!(resVal instanceof Hashtable)) {
            r = new SliceAuthorityReply<Hashtable>(res, null);
        } else {
            Hashtable userTable = (Hashtable) resVal;
            r = new SliceAuthorityReply<Hashtable>(res, userTable);
        }
        log(res, r, "resolveUser", "Resolve", con, methodParams);
        return r;
    }

    /**
     * Resolve call, with a any types: Get info about an object of the specified type.
     *
     * @param con            the {@code GeniConnection} to use for this call. Must be a connection to a slice authority.
     * @param userCredential the credential for the user requesting the info
     * @param type           a user ID
     * @param target         a user ID
     * @return a Hashtable containing info about the user, wrapped in a {@code SliceAuthority.SliceAuthorityReply}
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    @ApiMethod(order = 14, hint = "Resolve call, with a free URN and type arguments: Get info about a specified object.")
    public SliceAuthorityReply<Hashtable> resolveAny(SfaConnection con,
                                                     @ApiMethodParameter(name = "userCredential", hint = "the credential for the user requesting the info")
                                                     AnyCredential userCredential,
                                                     @ApiMethodParameter(name = "target", hint = "an URN")
                                                     ResourceId target,
                                                     @ApiMethodParameter(name = "type", hint = "the type to resolve.")
                                                     String type) throws JFedException {
        Map<String, Object> methodParams =
                makeMethodParameters("userCredential", userCredential, "target", target, "type", type);

        Hashtable params = new Hashtable();
        params.put(target.getType(), target.getValue());
        params.put("credential", userCredential.getCredentialXml());
        params.put("type", type);

        Vector argv = new Vector();
        argv.setSize(1);
        argv.set(0, params);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "Resolve", argv, methodParams);

        SliceAuthorityReply<Hashtable> r = null;

        Object resVal = res.getResultValueObject();
        if (!(resVal instanceof Hashtable)) {
            r = new SliceAuthorityReply<Hashtable>(res, null);
        } else {
            Hashtable userTable = (Hashtable) resVal;
            r = new SliceAuthorityReply<Hashtable>(res, userTable);
        }
        log(res, r, "resolveUser", "Resolve", con, methodParams);
        return r;
    }

    /**
     * BindToSlice call: give another user permission to use a slice.
     *
     * @param con             the {@code GeniConnection} to use for this call. Must be a connection to a slice authority.
     * @param sliceCredential the credential for the slice to be shared
     * @param user            the ID of the user which will get permission to use this slice. (urn format: {@code urn:publicid:IDN+<AUTHORITY>+user+<USER>})
     * @return true if successful, wrapped in a {@code SliceAuthority.SliceAuthorityReply}
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    @ApiMethod(order = 20, hint = "BindToSlice call: give another user permission to use a slice.")
    public SliceAuthorityReply<Boolean> bindToSlice(SfaConnection con,
                                                    @ApiMethodParameter(name = "sliceCredential", hint = "the credential for the slice to be shared")
                                                    AnyCredential sliceCredential,
                                                    @ApiMethodParameter(name = "user", hint = "the ID of the user which will get permission to use this slice (typically the other user's URN).")
                                                    ResourceId user) throws JFedException {
        return bindToSlice(con, sliceCredential.toCredentialList(), true, user);

    }
    @ApiMethod(order = 20, hint = "BindToSlice call: give another user permission to use a slice.")
    public SliceAuthorityReply<Boolean> bindToSlice(SfaConnection con,
                                                    @ApiMethodParameter(name = "sliceCredentials", hint = "a list if credentials for the slice to be shared")
                                                    List<AnyCredential> sliceCredentials,
                                                    @ApiMethodParameter(name = "user", hint = "the ID of the user which will get permission to use this slice (typically the other user's URN).")
                                                    ResourceId user) throws JFedException {
        return bindToSlice(con, sliceCredentials, false, user);
    }

    public SliceAuthorityReply<Boolean> bindToSlice(SfaConnection con,
                                                    List<AnyCredential> sliceCredentials,
                                                    boolean forceSendSingleCredential /*if true, a single credential is sent instead of a list*/,
                                                    ResourceId user) throws JFedException {
        Map<String, Object> methodParams;
        if (forceSendSingleCredential)
            methodParams = makeMethodParameters("sliceCredential", sliceCredentials.get(0), "user", user);
        else
            methodParams = makeMethodParameters("sliceCredentials", sliceCredentials, "user", user);

        Vector argv = new Vector();

        Hashtable params = new Hashtable();
        params.put(user.getType(), user.getValue());
        if (forceSendSingleCredential) {
            params.put("credential", sliceCredentials.get(0).getCredentialXml());
        } else {
            Vector credentials = createCredentialsVector(sliceCredentials);
            params.put("credentials", credentials);
        }

        argv.setSize(1);
        argv.set(0, params);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "BindToSlice", argv, methodParams);

        SliceAuthorityReply<Boolean> r = null;
        try {
            Integer success = (Integer) res.getResultValueObject();
            r = new SliceAuthorityReply<Boolean>(res, (!success.equals(0)));
        } catch (Exception e) {
            log(res, null, "bindToSlice", "BindToSlice", con, methodParams);
            throw new JFedException("Error parsing BindToSlice result credential:" + e.getMessage(), e);
        }
        log(res, r, "bindToSlice", "BindToSlice", con, methodParams);
        return r;
    }

    /**
     * Register call: create a new slice.
     *
     * @param con            the {@code GeniConnection} to use for this call. Must be a connection to a slice authority.
     * @param userCredential the credential of the user creating this slice.
     * @param slice          the URN of the new slice (urn format: {@code urn:publicid:IDN+<AUTHORITY>+slice+<SLICENAME>})
     * @return the {@code AnyCredential} for the created slice, wrapped in a {@code SliceAuthority.SliceAuthorityReply}
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    @ApiMethod(order = 25, hint = "Register call: create a new slice.")
    public SliceAuthorityReply<AnyCredential> register(SfaConnection con,
                                                       @ApiMethodParameter(name = "userCredential", hint = "the credential of the user creating this slice.")
                                                       AnyCredential userCredential,
                                                       @ApiMethodParameter(name = "slice", hint = "the URN of the new slice (urn syntax: urn:publicid:IDN+<AUTHORITY>+slice+<SLICENAME>)")
                                                       ResourceUrn slice) throws JFedException {
        return register(con, userCredential.toCredentialList(), true, slice);
    }

    /**
     * Register call: create a new slice. (this variations allows multiple credentials)
     *
     * @param con            the {@code GeniConnection} to use for this call. Must be a connection to a slice authority.
     * @param userCredentials a list of credentials of the user creating this slice.
     * @param slice          the URN of the new slice (urn format: {@code urn:publicid:IDN+<AUTHORITY>+slice+<SLICENAME>})
     * @return the {@code AnyCredential} for the created slice, wrapped in a {@code SliceAuthority.SliceAuthorityReply}
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    @ApiMethod(order = 26, hint = "Register call: create a new slice.")
    public SliceAuthorityReply<AnyCredential> register(SfaConnection con,
                                                       @ApiMethodParameter(name = "userCredentials",
                                                               hint = "a list of credentials of the user creating this slice.",
                                                               parameterType = ApiMethodParameterType.LIST_OF_CREDENTIAL)
                                                       List<AnyCredential> userCredentials,
                                                       @ApiMethodParameter(name = "slice", hint = "the URN of the new slice (urn syntax: urn:publicid:IDN+<AUTHORITY>+slice+<SLICENAME>)")
                                                       ResourceUrn slice) throws JFedException {
        return register(con, userCredentials, false, slice);
    }

    public SliceAuthorityReply<AnyCredential> register(SfaConnection con,
                                                       List<AnyCredential> userCredentials,
                                                       boolean forceSendSingleCredential /*if true, a single credential is sent instead of a list*/,
                                                       ResourceUrn slice) throws JFedException {
        assert con != null;
        assert userCredentials != null;
        assert !userCredentials.isEmpty();
        assert slice != null;

        Map<String, Object> methodParams;
        if (forceSendSingleCredential) {
            assert !userCredentials.isEmpty();
            methodParams = makeMethodParameters("userCredential", userCredentials.get(0), "slice", slice);
        } else
            methodParams = makeMethodParameters("userCredentials", userCredentials, "slice", slice);

        Vector argv = new Vector();

        if (!slice.getType().equals("urn"))
            throw new JFedException("Bad type for ResourceId slice (\"" + slice.getType() + "\"). Only \"urn\" is supported here");

        Hashtable params = new Hashtable();
        params.put("type", "Slice");
        if (forceSendSingleCredential) {
            params.put("credential", userCredentials.get(0).getCredentialXml());
        } else {
            Vector credentials = createCredentialsVector(userCredentials);
            params.put("credentials", credentials);
        }
        params.put(slice.getType(), slice.getValue());

        argv.setSize(1);
        argv.set(0, params);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "Register", argv, methodParams);

        SliceAuthorityReply<AnyCredential> r = null;
        int intCode = (Integer) res.getResultCode();
        GeniResponseCode genicode = GeniAMResponseCode.getByCode(intCode);
        if (genicode.isSuccess()) {
            try {
                AnyCredential sliceCredential =
                        AnyCredential.createSfa2("SliceAuthority register " + slice.getValue(), res.getResultValueObject().toString());
                r = new SliceAuthorityReply<AnyCredential>(res, sliceCredential);
            } catch (Exception e) {
                log(res, null, "register", "Register", con, methodParams);
                throw new JFedException("Error parsing BindToSlice result credential:" + e.getMessage(), e);
            }
        } else
            r = new SliceAuthorityReply<AnyCredential>(res, null);

        log(res, r, "register", "Register", con, methodParams);
        return r;
    }

    /**
     * RenewSlice call: Change the expiration date of a slice.
     *
     * @param con                the {@code GeniConnection} to use for this call. Must be a connection to a slice authority.
     * @param sliceCredential    the credential for the slice to be renewed
     * @param expiration_rfc3339 a string with the new expiration date, in RFC3339 format. See {@link be.iminds.ilabt.jfed.util.RFC3339Util} for help converting to the correct format.
     * @return the {@code AnyCredential} for the renewed slice, wrapped in a {@code SliceAuthority.SliceAuthorityReply}
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     * @see <a href="https://tools.ietf.org/html/rfc3339">RFC 3339</a>
     */
    @ApiMethod(order = 30, hint = "RenewSlice call: Change the expiration date of a slice.")
    public SliceAuthorityReply<AnyCredential> renewSlice(SfaConnection con,
                                                         @ApiMethodParameter(name = "sliceCredential", hint = "the credential for the slice to be renewed")
                                                         AnyCredential sliceCredential,
                                                         @ApiMethodParameter(name = "expiration_rfc3339", hint = "A string with the new experiation date, in RFC3339 format.")
                                                         String expiration_rfc3339) throws JFedException {
        return renewSlice(con, sliceCredential.toCredentialList(), true, expiration_rfc3339);
    }
    @ApiMethod(order = 31, hint = "RenewSlice call: Change the expiration date of a slice.")
    public SliceAuthorityReply<AnyCredential> renewSlice(SfaConnection con,
                                                         @ApiMethodParameter(name = "sliceCredentials",
                                                                 hint = "the credential for the slice to be renewed",
                                                                 parameterType = ApiMethodParameterType.LIST_OF_CREDENTIAL)
                                                         List<AnyCredential> sliceCredentials,
                                                         @ApiMethodParameter(name = "expiration_rfc3339", hint = "A string with the new experiation date, in RFC3339 format.")
                                                         String expiration_rfc3339) throws JFedException {
        return renewSlice(con, sliceCredentials, false, expiration_rfc3339);
    }

    public SliceAuthorityReply<AnyCredential> renewSlice(SfaConnection con,
                                                         @ApiMethodParameter(name = "sliceCredentials",
                                                                 hint = "the credential for the slice to be renewed",
                                                                 parameterType = ApiMethodParameterType.LIST_OF_CREDENTIAL)
                                                         List<AnyCredential> sliceCredentials,
                                                         boolean forceSendSingleCredential /*if true, a single credential is sent instead of a list*/,
                                                         String expiration_rfc3339) throws JFedException {
        Map<String, Object> methodParams;
        if (forceSendSingleCredential) {
            assert !sliceCredentials.isEmpty();
            methodParams =
                    makeMethodParameters("sliceCredential", sliceCredentials.get(0), "expiration_rfc3339", expiration_rfc3339);
        } else {
            methodParams =
                    makeMethodParameters("sliceCredentials", sliceCredentials, "expiration_rfc3339", expiration_rfc3339);
        }
        Vector argv = new Vector();

        Hashtable params = new Hashtable();
        params.put("expiration", expiration_rfc3339);
        if (forceSendSingleCredential) {
            params.put("credential", sliceCredentials.get(0).getCredentialXml());
        } else {
            Vector credentials = createCredentialsVector(sliceCredentials);
            params.put("credentials", credentials);
        }

        argv.setSize(1);
        argv.set(0, params);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "RenewSlice", argv, methodParams);

        SliceAuthorityReply<AnyCredential> r = null;
        int intCode = (Integer) res.getResultCode();
        GeniResponseCode genicode = GeniAMResponseCode.getByCode(intCode);
        if (genicode.isSuccess()) {
            try {
                AnyCredential newSliceCredential =
                        AnyCredential.createSfa2("SliceAuthority renewSlice", res.getResultValueObject().toString());
                r = new SliceAuthorityReply<AnyCredential>(res, newSliceCredential);
            } catch (Exception e) {
                log(res, null, "renewSlice", "RenewSlice", con, methodParams);
                throw new JFedException("Error parsing RenewSlice result credential:" + e.getMessage(), e);
            }
        } else
            r = new SliceAuthorityReply<AnyCredential>(res, null);

        log(res, r, "renewSlice", "RenewSlice", con, methodParams);
        return r;
    }

    /**
     * <p>Shutdown call:
     * <p/>
     * <p>API description:
     * <p>Perform an emergency shutdown on a slice, by asking the SA (for that slice) to do an emergency shutdown. Operationally, the request is forwarded to the ClearingHouse which knows the full set of Component Managers. The call returns once the ClearingHouse is notified; the ClearingHouse will process the request asynchronously.
     * <p/>
     * <p/>
     * <p/>
     * <p>Clearinghouse API description:
     * <p>Perform an emergency shutdown on a slice. This is typically invoked by the Slice Authority for the slice, but may be invoked by anyone with a clearinghouse credential (this needs to change to allow anyone with a valid slice credential). As it stands, anyone with a slice credential can contact the Slice Authority for the slice, and ask it to do the shutown operation.
     * <p>Since the Clearinghouse must contact each Component Manager to tell it to shutdown the slice, this call will return immediately. There is currently no facility to find out if/when the shutdown has completed.
     *
     * @param con             the {@code GeniConnection} to use for this call. Must be a connection to a slice authority.
     * @param sliceCredential the credential for the slice to be shut down
     * @return a boolean (success or failure), wrapped in a {@code SliceAuthority.SliceAuthorityReply}
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    @ApiMethod(order = 40, hint = "API description:\n" +
            "Perform an emergency shutdown on a slice, by asking the SA (for that slice) to do an emergency shutdown. Operationally, the request is forwarded to the ClearingHouse which knows the full set of Component Managers. The call returns once the ClearingHouse is notified; the ClearingHouse will process the request asynchronously.\n" +
            "\n" +
            "Clearinghouse API description:\n" +
            "Perform an emergency shutdown on a slice. This is typically invoked by the Slice Authority for the slice, but may be invoked by anyone with a clearinghouse credential (this needs to change to allow anyone with a valid slice credential). As it stands, anyone with a slice credential can contact the Slice Authority for the slice, and ask it to do the shutown operation.\n" +
            "Since the Clearinghouse must contact each Component Manager to tell it to shutdown the slice, this call will return immediately. There is currently no facility to find out if/when the shutdown has completed.\n")
    public SliceAuthorityReply<Boolean> shutdown(SfaConnection con,
                                                 @ApiMethodParameter(name = "sliceCredential", hint = "The credential of the slice to shut down.")
                                                 AnyCredential sliceCredential) throws JFedException {
        Map<String, Object> methodParams = makeMethodParameters("sliceCredential", sliceCredential);

        Vector argv = new Vector();

        Hashtable params = new Hashtable();
        params.put("credential", sliceCredential.getCredentialXml());

        argv.setSize(1);
        argv.set(0, params);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "Shutdown", argv, methodParams);

        SliceAuthorityReply<Boolean> r = new SliceAuthorityReply<Boolean>(res);
        log(res, r, "shutdown", "Shutdown", con, methodParams);
        return r;
    }

    /**
     * GetKeys call: Get the public SSH key()s of a user.
     *
     * @param con            the {@code GeniConnection} to use for this call. Must be a connection to a slice authority.
     * @param userCredential the credential of the user retrieving ssh keys
     * @return a list of keys, wrapped in a {@code SliceAuthority.SliceAuthorityReply}
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    @ApiMethod(order = 50, hint = "GetKeys call: Get the public SSH key()s of a user.")
    public SliceAuthorityReply<List<String>> getKeys(SfaConnection con,
                                                     @ApiMethodParameter(name = "userCredential", hint = "the credential of the user retrieving ssh keys")
                                                     AnyCredential userCredential) throws JFedException {
        Map<String, Object> methodParams = makeMethodParameters("userCredential", userCredential);

        Vector argv = new Vector();

        Hashtable params = new Hashtable();
        params.put("credential", userCredential.getCredentialXml());

        argv.setSize(1);
        argv.set(0, params);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "GetKeys", argv, methodParams);

        List<String> keyList = new ArrayList<String>();
        try {
            Vector keys = (Vector) res.getResultValueObject();
            for (Object key : keys) {
                if (key instanceof String)
                    keyList.add((String) key);
                else {
                    if (key instanceof Hashtable) {
                        Hashtable t = (Hashtable) key;
                        String type = (String) t.get("type");
                        if (!type.equals("ssh"))
                            LOG.warn("Key is hashtable with non 'ssh' type: " + t);
                        keyList.add((String) t.get("key"));
                    } else
                        throw new JFedException("key is of type " + key.getClass().getName() + " val=" + key.toString());
                }
            }
        } catch (Exception e) {
            LOG.warn("Warning (ignored): getKeys did not return a Vector<string> but a " + res.getResultValueObject().getClass() + " -> " + e.getMessage() + " res.getResultValueObject()=" + res.getResultValueObject(), e);
        }
        SliceAuthorityReply<List<String>> r = new SliceAuthorityReply<List<String>>(res, keyList);
        log(res, r, "getKeys", "GetKeys", con, methodParams);
        return r;
    }

    /**
     * Remove call: remove a slice.
     * <p/>
     * Note: at the time of writing this documentation, remove was not implemented at the server.
     *
     * @param con            the {@code GeniConnection} to use for this call. Must be a connection to a slice authority.
     * @param userCredential the credential of the user removing a slice.
     * @param slice          the ID of the slice to be removed.
     * @return a boolean (success or failure), wrapped in a {@code SliceAuthority.SliceAuthorityReply}
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    @ApiMethod(order = 60, hint = " Remove call: remove a slice.\n\nNote: at the time of writing this documentation, remove was not implemented at the server.")
    public SliceAuthorityReply<Boolean> remove(SfaConnection con,
                                               @ApiMethodParameter(name = "userCredential", hint = "The credential of the user removing a slice.")
                                               AnyCredential userCredential,
                                               @ApiMethodParameter(name = "slice", hint = "The ID of the slice to be removed (typically the slice URN).")
                                               ResourceId slice) throws JFedException {
        Map<String, Object> methodParams = makeMethodParameters("userCredential", userCredential, "slice", slice);

        Vector argv = new Vector();

        Hashtable params = new Hashtable();
        params.put("type", "Slice");
        params.put("credential", userCredential.getCredentialXml());
        params.put(slice.getType(), slice.getValue());

        argv.setSize(1);
        argv.set(0, params);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "Remove", argv, methodParams);

        SliceAuthorityReply<Boolean> r = new SliceAuthorityReply<Boolean>(res);
        log(res, r, "remove", "Remove", con, methodParams);
        return r;
    }

    public static class SliceAuthorityReply<T> implements SfaApiCallReply<T> {
        private final GeniAMResponseCode genicode;
        private final T val;
        /* output is typically set only on error*/
        private final String output;

        private final Hashtable rawResult;
        private final XMLRPCCallDetailsWithCodeValueError xmlrpcCallDetails;

        public SliceAuthorityReply(XMLRPCCallDetailsWithCodeValueError res) {
            this.xmlrpcCallDetails = res;
            this.rawResult = res.getResult();

            int intCode = (Integer) res.getResultCode();
            this.genicode = GeniAMResponseCode.getByCode(intCode);
            T v;
            try {
                v = (T) res.getResultValueObject();
            } catch (Exception e) {
                v = null;
            }
            this.val = v;
            this.output = res.getResultOutput();
        }


        public SliceAuthorityReply(XMLRPCCallDetailsWithCodeValueError res, T val) {
            this.xmlrpcCallDetails = res;
            this.rawResult = res.getResult();

            int intCode = (Integer) res.getResultCode();
            this.genicode = GeniAMResponseCode.getByCode(intCode);
            this.val = val;
            this.output = res.getResultOutput();
        }

        public int getCode() {
            return genicode.getCode();
        }

        public GeniAMResponseCode getGeniResponseCode() {
            return genicode;
        }

        public Hashtable getRawResult() {
            return rawResult;
        }

        public T getValue() {
            return val;
        }

        public String getOutput() {
            return output;
        }

        @Override public XMLRPCCallDetailsWithCodeValueError getXMLRPCCallDetailsWithCodeValueError() {
            return xmlrpcCallDetails;
        }
        @Override public XMLRPCCallDetails getXMLRPCCallDetails() { return xmlrpcCallDetails; }
    }


    //TODO: add DiscoverResources  (ignored because same as ClearingHouse ListComponents)
}
